from __future__ import annotations

from datetime import timedelta

from momento.retry import FixedCountRetryStrategy

from .configuration import Configuration
from .transport.transport_strategy import (
    StaticGrpcConfiguration,
    StaticTransportStrategy,
)


class Configurations:
    """Container class for pre-built configurations."""

    class Laptop(Configuration):
        """Laptop config provides defaults suitable for a medium-to-high-latency dev environment.

        Permissive timeouts, retries, and relaxed latency and throughput targets.
        """

        @staticmethod
        def latest() -> Configurations.Laptop:
            """Provides the latest recommended configuration for a laptop development environment.

            This configuration will be updated every time there is a new version of the laptop configuration.
            """
            return Configurations.Laptop.v1()

        @staticmethod
        def v1() -> Configurations.Laptop:
            """Provides the v1 recommended configuration for a laptop development environment.

            This configuration is guaranteed not to change in future releases of the Momento Python SDK.
            """
            return Configurations.Laptop(
                StaticTransportStrategy(StaticGrpcConfiguration(timedelta(seconds=15))),
                FixedCountRetryStrategy(max_attempts=3),
            )

    class InRegion:
        """Default for application running in the same region as the Momento service.

        InRegion provides defaults suitable for an environment where your client is running in the same region as the
        Momento service.  It has more aggressive timeouts and retry behavior than the Laptop config.
        """

        class Default(Configuration):
            """Prioritizes throughput and client resource utilization.

            It has a slightly relaxed client-side timeout setting to maximize throughput.
            """

            @staticmethod
            def latest() -> Configurations.InRegion.Default:
                """Provides the latest recommended configuration for a typical in-region environment.

                This configuration will be updated every time there is a new version of the in-region configuration.
                """
                return Configurations.InRegion.Default.v1()

            @staticmethod
            def v1() -> Configurations.InRegion.Default:
                """Provides the v1 recommended configuration for a typical in-region environment.

                This configuration is guaranteed not to change in future releases of the Momento Python SDK.
                """
                return Configurations.InRegion.Default(
                    StaticTransportStrategy(StaticGrpcConfiguration(timedelta(milliseconds=1100))),
                    FixedCountRetryStrategy(max_attempts=3),
                )

        class LowLatency(Configuration):
            """Prioritizes keeping p99.9 latencies as low as possible.

            It potentially sacrifices some throughput to achieve this.  It has a very aggressive client-side timeout.
            Use this configuration if the most important factor is to ensure that cache unavailability doesn't
            force unacceptably high latencies for your application.
            """

            @staticmethod
            def latest() -> Configurations.InRegion.LowLatency:
                """Provides the latest recommended configuration for an environment with low-latency requirements.

                This configuration will be updated every time there is a new version of the low-latency configuration.
                """
                return Configurations.InRegion.LowLatency.v1()

            @staticmethod
            def v1() -> Configurations.InRegion.LowLatency:
                """Provides the v1 recommended configuration for an environment with low-latency requirements.

                This configuration is guaranteed not to change in future releases of the Momento Python SDK.
                """
                return Configurations.InRegion.LowLatency(
                    StaticTransportStrategy(StaticGrpcConfiguration(timedelta(milliseconds=500))),
                    FixedCountRetryStrategy(max_attempts=3),
                )



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass

from momento.responses.response import ControlResponse

from ...mixins import ErrorResponseMixin


class RevokeSigningKeyResponse(ControlResponse):
    """Parent response type for a cache `revoke_signing_key` request.

    Its subtypes are:
    - `RevokeSigningKey.Success`
    - `RevokeSigningKey.Error`

    See `CacheClient` for how to work with responses.
    """


class RevokeSigningKey(ABC):
    """Groups all `RevokeSigningKeyResponse` derived types under a common namespace."""

    @dataclass
    class Success(RevokeSigningKeyResponse):
        """The response from revoking a signing key."""

    class Error(RevokeSigningKeyResponse, ErrorResponseMixin):
        """Contains information about an error returned from a request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """



from __future__ import annotations

import collections.abc
from datetime import timedelta
from typing import Iterable, Optional, Tuple

from momento.errors import InvalidArgumentException
from momento.typing import (
    TDictionaryFields,
    TDictionaryItems,
    TListValuesInput,
    TSetElementsInput,
    TSetElementsInputBytes,
)

DEFAULT_BYTES_CONVERSION_ERROR = "Could not convert the given type to bytes: "
DEFAULT_LIST_CONVERSION_ERROR = "The given type is not list[str | bytes]: "
DEFAULT_DICTIONARY_CONVERSION_ERROR = "The given type is not a valid Mapping: "
DEFAULT_DICTIONARY_FIELDS_CONVERSION_ERROR = "The given type is not Iterable[str | bytes]: "
DEFAULT_SET_CONVERSION_ERROR = "The given type is not set[str | bytes]: "


def _validate_name(name: str, field_name: str) -> None:
    if not isinstance(name, str):
        raise InvalidArgumentException(f"{field_name} must be a string")
    if name == "":
        raise InvalidArgumentException(f"{field_name} must not be empty")


def _validate_cache_name(cache_name: str) -> None:
    _validate_name(cache_name, "Cache name")


def _validate_list_name(list_name: str) -> None:
    _validate_name(list_name, "List name")


def _validate_dictionary_name(dictionary_name: str) -> None:
    _validate_name(dictionary_name, "Dictionary name")


def _validate_set_name(set_name: str) -> None:
    _validate_name(set_name, "Set name")


def _as_bytes(
    data: str | bytes,
    error_message: Optional[str] = DEFAULT_BYTES_CONVERSION_ERROR,
) -> bytes:
    if isinstance(data, str):
        return data.encode("utf-8")
    if isinstance(data, bytes):
        return data
    raise InvalidArgumentException(f"{error_message}{type(data)}")


def _gen_iterable_as_bytes(values: Iterable[str | bytes], error_message: str) -> Iterable[bytes]:
    if not isinstance(values, collections.abc.Iterable):
        raise InvalidArgumentException(f"{error_message}{type(values)}")
    for value in values:
        yield _as_bytes(value)


def _gen_list_as_bytes(values: TListValuesInput, error_message: str = DEFAULT_LIST_CONVERSION_ERROR) -> Iterable[bytes]:
    yield from _gen_iterable_as_bytes(values, error_message)


def _gen_dictionary_items_as_bytes(
    items: TDictionaryItems, error_message: str = DEFAULT_DICTIONARY_CONVERSION_ERROR
) -> Iterable[Tuple[bytes, bytes]]:
    if not isinstance(items, collections.abc.Mapping):
        raise InvalidArgumentException(f"{error_message}{type(items)}")
    for key, value in items.items():
        yield (_as_bytes(key), _as_bytes(value))


def _gen_dictionary_fields_as_bytes(
    fields: TDictionaryFields, error_message: str = DEFAULT_DICTIONARY_FIELDS_CONVERSION_ERROR
) -> Iterable[bytes]:
    yield from _gen_iterable_as_bytes(fields, error_message)


def _gen_set_input_as_bytes(
    elements: TSetElementsInput, error_message: str = DEFAULT_SET_CONVERSION_ERROR
) -> TSetElementsInputBytes:
    # NB: the set input does not need to be unique
    yield from _gen_iterable_as_bytes(elements, error_message)


def _validate_timedelta_ttl(ttl: Optional[timedelta], field_name: str) -> None:
    if not isinstance(ttl, timedelta):
        raise InvalidArgumentException(f"{field_name} must be a timedelta.")
    if ttl.total_seconds() <= 0:
        raise InvalidArgumentException(f"{field_name} must be a positive amount of time.")


def _validate_ttl(ttl: Optional[timedelta]) -> None:
    _validate_timedelta_ttl(ttl=ttl, field_name="TTL")


def _validate_request_timeout(request_timeout: Optional[timedelta]) -> None:
    if request_timeout is None:
        return
    _validate_timedelta_ttl(ttl=request_timeout, field_name="Request timeout")



from .default_eligibility_strategy import DefaultEligibilityStrategy
from .eligibility_strategy import EligibilityStrategy
from .fixed_count_retry_strategy import FixedCountRetryStrategy
from .retry_strategy import RetryStrategy
from .retryable_props import RetryableProps

__all__ = [
    "DefaultEligibilityStrategy",
    "EligibilityStrategy",
    "FixedCountRetryStrategy",
    "RetryStrategy",
    "RetryableProps",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta

from momento.retry import RetryStrategy

from .transport.transport_strategy import TransportStrategy


class ConfigurationBase(ABC):

    # TODO: Middlewares
    @abstractmethod
    def get_retry_strategy(self) -> RetryStrategy:
        pass

    @abstractmethod
    def with_retry_strategy(self, retry_strategy: RetryStrategy) -> Configuration:
        pass

    @abstractmethod
    def get_transport_strategy(self) -> TransportStrategy:
        pass

    @abstractmethod
    def with_transport_strategy(self, transport_strategy: TransportStrategy) -> Configuration:
        pass

    @abstractmethod
    def with_client_timeout(self, client_timeout: timedelta) -> Configuration:
        pass


class Configuration(ConfigurationBase):
    """Configuration options for Momento Simple Cache Client."""

    def __init__(self, transport_strategy: TransportStrategy, retry_strategy: RetryStrategy):
        """Instantiate a Configuration.

        Args:
            transport_strategy (TransportStrategy): Configuration options for networking with
            the Momento service.
            retry_strategy (RetryStrategy): the strategy to use when determining whether to retry a grpc call.
        """
        self._transport_strategy = transport_strategy
        self._retry_strategy = retry_strategy

    def get_retry_strategy(self) -> RetryStrategy:
        """Access the retry strategy.

        Returns:
            RetryStrategy: the strategy to use when determining whether to retry a grpc call.
        """
        return self._retry_strategy

    def with_retry_strategy(self, retry_strategy: RetryStrategy) -> Configuration:
        """Copy constructor for overriding RetryStrategy.

        Args:
            retry_strategy (RetryStrategy): the new RetryStrategy.

        Returns:
            Configuration: the new Configuration with the specified RetryStrategy.
        """
        return Configuration(self._transport_strategy, retry_strategy)

    def get_transport_strategy(self) -> TransportStrategy:
        """Access the transport strategy.

        Returns:
            TransportStrategy: the current configuration options for wire interactions with the Momento service.
        """
        return self._transport_strategy

    def with_transport_strategy(self, transport_strategy: TransportStrategy) -> Configuration:
        """Copy constructor for overriding TransportStrategy.

        Args:
            transport_strategy (TransportStrategy): the new TransportStrategy.

        Returns:
            Configuration: the new Configuration with the specified TransportStrategy.
        """
        return Configuration(transport_strategy, self._retry_strategy)

    def with_client_timeout(self, client_timeout: timedelta) -> Configuration:
        """Copies the Configuration and sets the new client-side timeout in the copy's TransportStrategy.

        Args:
            client_timeout (timedelta): the new client-side timeout.

        Return:
            Configuration: the new Configuration.
        """
        return Configuration(self._transport_strategy.with_client_timeout(client_timeout), self._retry_strategy)



from __future__ import annotations

import grpc

from .eligibility_strategy import EligibilityStrategy
from .retryable_props import RetryableProps

RETRYABLE_STATUS_CODES: list[grpc.StatusCode] = [
    # # including all the status codes for reference, but
    # # commenting out the ones we don't want to retry on for now.
    # grpc.StatusCode.OK,
    # grpc.StatusCode.CANCELLED,
    # grpc.StatusCode.UNKNOWN,
    # grpc.StatusCode.INVALID_ARGUMENT,
    # grpc.StatusCode.DEADLINE_EXCEEDED,
    # grpc.StatusCode.NOT_FOUND,
    # grpc.StatusCode.ALREADY_EXISTS,
    # grpc.StatusCode.PERMISSION_DENIED,
    # grpc.StatusCode.RESOURCE_EXHAUSTED,
    # grpc.StatusCode.FAILED_PRECONDITION,
    # grpc.StatusCode.ABORTED,
    # grpc.StatusCode.OUT_OF_RANGE,
    # grpc.StatusCode.UNIMPLEMENTED,
    grpc.StatusCode.INTERNAL,  # type: ignore[misc]
    grpc.StatusCode.UNAVAILABLE,  # type: ignore[misc]
    # grpc.StatusCode.DATA_LOSS,
    # grpc.StatusCode.UNAUTHENTICATED,
]

RETRYABLE_REQUEST_TYPES: list[str] = [
    "/cache_client.Scs/Set",
    "/cache_client.Scs/Get",
    "/cache_client.Scs/Delete",
    # not idempotent: "/cache_client.Scs/Increment"
    "/cache_client.Scs/DictionarySet",
    # not idempotent: "/cache_client.Scs/DictionaryIncrement",
    "/cache_client.Scs/DictionaryGet",
    "/cache_client.Scs/DictionaryFetch",
    "/cache_client.Scs/DictionaryDelete",
    "/cache_client.Scs/SetUnion",
    "/cache_client.Scs/SetDifference",
    "/cache_client.Scs/SetFetch",
    # not idempotent: "/cache_client.Scs/SetIfNotExists"
    # not idempotent: "/cache_client.Scs/ListPushFront",
    # not idempotent: "/cache_client.Scs/ListPushBack",
    # not idempotent: "/cache_client.Scs/ListPopFront",
    # not idempotent: "/cache_client.Scs/ListPopBack",
    "/cache_client.Scs/ListFetch",
    # Warning: in the future, this may not be idempotent
    # Currently it supports removing all occurrences of a value.
    # In the future, we may also add "the first/last N occurrences of a value".
    # In the latter case it is not idempotent.
    "/cache_client.Scs/ListRemove",
    "/cache_client.Scs/ListLength",
    # not idempotent: "/cache_client.Scs/ListConcatenateFront",
    # not idempotent: "/cache_client.Scs/ListConcatenateBack"
]


class DefaultEligibilityStrategy(EligibilityStrategy):
    def is_eligible_for_retry(self, props: RetryableProps) -> bool:
        """Determines whether a grpc call is able to be retried.

        The determination is based on the result of the last invocation of the call and whether the call is idempotent.

        Args:
            props (RetryableProps): Information about the grpc call and its last invocation.
        """
        if props.grpc_status not in RETRYABLE_STATUS_CODES:  # type: ignore[misc]
            return False

        if props.grpc_method not in RETRYABLE_REQUEST_TYPES:
            return False

        return True



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Optional

from momento.internal._utilities import _validate_timedelta_ttl


@dataclass
class CollectionTtl:
    """Represents the desired behavior for managing the TTL on collection objects in your cache.

    For cache operations that modify a collection (dictionaries, lists, sets, etc), there are a few things
    to consider.  The first time the collection is created, we need to
    set a TTL on it.  For subsequent operations that modify the collection
    you may choose to update the TTL in order to prolong the life of the
    cached collection object, or you may choose to leave the TTL unmodified
    in order to ensure that the collection expires at the original TTL.

    The default behavior is to refresh the TTL (to prolong the life of the
    collection) each time it is written.  This behavior can be modified
    by calling the method `with_no_refresh_ttl_on_updates`.

    Raises:
        ValueError: if the ttl is not a positive amount of time.
    """

    ttl: Optional[timedelta] = None
    """The duration after which the cached collection should be expired from the cache.
    If `null`, we use the default TTL that was passed to the `CacheClient` init method.
    """
    refresh_ttl: bool = True
    """If `True`, the collection's TTL will be refreshed (to prolong the life of the collection)
    on every update.  If `False`, the collection's TTL will only be set when the collection is
    initially created.
    """

    def __post_init__(self) -> None:
        if self.ttl is None:
            return
        _validate_timedelta_ttl(ttl=self.ttl, field_name="ttl")

    @staticmethod
    def from_cache_ttl() -> CollectionTtl:
        """The default way to handle TTLs for collections.

        The default TTL `timedelta` that was specified when instantiating the
        `CacheClient` will be used, and the TTL for the collection will be
        refreshed any time the collection is modified.

        Returns:
            CollectionTtl
        """
        return CollectionTtl(ttl=None, refresh_ttl=True)

    @staticmethod
    def of(ttl: timedelta) -> CollectionTtl:
        """Constructs a CollectionTtl with the specified `timedelta`.

        TTL for the collection will be refreshed any time the collection is
        modified.

        Args:
            ttl (timedelta): the duration to use for the TTL

        Returns:
            CollectionTtl
        """
        return CollectionTtl(ttl=ttl)

    @staticmethod
    def refresh_ttl_if_provided(ttl: Optional[timedelta] = None) -> CollectionTtl:
        """Constructs a `CollectionTtl` with the specified `timedelta`.

        Will only refresh if the TTL is provided (ie not `null`).

        Args:
            ttl (Optional[timedelta], optional): the duration to use for the TTL. Defaults to None.

        Returns:
            CollectionTtl
        """
        return CollectionTtl(ttl=ttl, refresh_ttl=ttl is not None)

    def with_refresh_ttl_on_updates(self) -> CollectionTtl:
        """Specifies the TTL for the collection be refreshed when the collection is modified.

        This is the default behavior for a CollectionTTL.

        Returns:
            CollectionTtl
        """
        return CollectionTtl(ttl=self.ttl, refresh_ttl=True)

    def with_no_refresh_ttl_on_updates(self) -> CollectionTtl:
        """Specifies the TTL for the collection should not be refreshed when the collection is modified.

        Use this to ensure your collection expires at the originally specified time,
        even if you make modifications to the collection.

        Returns:
            CollectionTtl
        """
        return CollectionTtl(ttl=self.ttl, refresh_ttl=False)



from abc import ABC

from ...mixins import ErrorResponseMixin
from ...response import CacheResponse


class CacheListRemoveValueResponse(CacheResponse):
    """Response type for a `list_remove_value` request.

    Its subtypes are:
    - `CacheListRemoveValue.Success`
    - `CacheListRemoveValue.Error`

    See `CacheClient` for how to work with responses.
    """


class CacheListRemoveValue(ABC):
    """Groups all `CacheListRemoveValueResponse` derived types under a common namespace."""

    class Success(CacheListRemoveValueResponse):
        """Indicates removing the values was successful."""

    class Error(CacheListRemoveValueResponse, ErrorResponseMixin):
        """Indicates an error occured in the request.

        This includes:
        - `error_code`: `MomentoErrorCode` value for the error.
        - `messsage`: a detailed error message.
        """
